import json

import arviz as az
import numpy as np
from sklearn.preprocessing import MinMaxScaler
from matplotlib import pyplot as plt

from pymob.simulation import SimulationBase
from pymob.inference.base import InferenceBackend, PymobInferenceData
from pymob.utils.errors import import_optional_dependency


extra = "'pymoo' dependencies can be installed with pip install pymob[pymoo]"
pymoo = import_optional_dependency("pymoo", errors="warn", extra=extra)
if pymoo is not None:
    import pathos.multiprocessing as mp
    from pymoo.algorithms.moo.unsga3 import UNSGA3
    from pymoo.core.problem import ElementwiseProblem, Problem
    from pymoo.util.ref_dirs import get_reference_directions
    from pymoo.termination.default import DefaultMultiObjectiveTermination


class PymooBackend:
    idata: PymobInferenceData
    def __init__(
        self, 
        simulation: SimulationBase,
    ):
        self.simulation = simulation
        self.config = self.simulation.config
        self.pool: mp.Pool = None  # type: ignore
        self.distance_function = self.distance_function_parser()
        self.transform = self.variable_parser()
        self.problem = OptimizationProblem(backend=self)

    
    def distance_function_parser(self):
        def f(x):
            evaluator = self.simulation.dispatch(theta=x)
            evaluator()
            obj_name, obj_value = self.simulation.objective_function(
                results=evaluator.results
            )
            return obj_value
        return f

    def variable_mapper(self, x):
        names = self.simulation.model_parameter_names
        return {n:x_i for n, x_i in zip(names, x)}

    def variable_parser(self):
        variables = self.simulation.free_model_parameters

        bounds = []
        names = []
        for v in variables:
            if v.min is None or v.max is None:
                raise ValueError(
                    f"Bounds are not fully defined in Param({v}). Bounds (min, max) "
                    "must be defined when using the pymoo backend.")
            bounds.append([v.min, v.max])
            names.append(v.name)

        bounds = np.array(bounds).T
        scaler = MinMaxScaler().fit(bounds)

        # check that parameter values and names match, so variable_mapper 
        # can be safely called
        assert names == self.simulation.model_parameter_names

        def transform(X_scaled):
            X = scaler.inverse_transform(X_scaled)
            return map(self.variable_mapper, X)

        return transform
    
    def store_results(self, results):
        params = list(self.transform(X_scaled=np.array(results.X, ndmin=2)) )[0]

        res_dict = {
            "f": results.f,
            "cv": results.cv,
            "X": params
        }
        
        file = f"{self.config.case_study.output_path}/pymoo_params.json"
        with open(file, "w") as fp:
            json.dump(res_dict, fp, indent=4)

        print(f"written results to {file}")

    def run(self):
        """Implements the parallelization in pymoo"""
        n_cores = self.config.multiprocessing.n_cores
        print(f"Using {n_cores} CPU cores")

        if n_cores == 1:
            self.optimize()
        elif n_cores > 1:
            with mp.ProcessingPool(n_cores) as pool:
                self.pool = pool  # type: ignore # assign pool so it can be accessed by _evaluate
                self.optimize()


    def optimize(self):

        reference_directions = get_reference_directions(
            name="energy", 
            n_dim=self.config.inference.n_objectives, 
            n_points=self.config.inference_pymoo.population_size,
            seed=self.config.simulation.seed + 1
        )

        algorithm = UNSGA3(
            ref_dirs=reference_directions,
            pop_size=self.config.inference_pymoo.population_size
        )

        termination = DefaultMultiObjectiveTermination(
            xtol=self.config.inference_pymoo.xtol,
            cvtol=self.config.inference_pymoo.cvtol,
            ftol=self.config.inference_pymoo.ftol,
            n_max_gen=self.config.inference_pymoo.max_nr_populations,
        )

        # prepare the algorithm to solve the specific problem (same arguments as for the minimize function)
        algorithm.setup(
            problem=self.problem, 
            termination=termination,
            seed=self.config.simulation.seed, 
            verbose=True,
        )


        # until the algorithm has no terminated
        while algorithm.has_next():

            # ask the algorithm for the next solution to be evaluated
            pop = algorithm.ask()

            # evaluate the individuals using the algorithm's evaluator (necessary to count evaluations for termination)
            algorithm.evaluator.eval(self.problem, pop) # type: ignore

            # returned the evaluated individuals which have been evaluated or even modified
            algorithm.tell(infills=pop)

            # self.post_processing(pop=pop)

        # obtain the result objective from the algorithm
        res = algorithm.result()

        # save the results
        self.results = res
        # TODO: Store results to idata (although it seems pointless),
        # extra dimensions for individuals and generations, one draw nevertheless

        self.store_results(results=res)

    def post_processing(self, pop):
        F = pop.get("F")
        X = pop.get("X")

        f_min = F.min()
        x_min = X[np.where(F[:,0] == f_min)]
        # print(
        #     f"Generation {gen}: f_min={f_min}, x_min={x_min}",
        #     flush=True
        # )

    def load_results(self):
        with open(f"{self.config.case_study.output_path}/pymoo_params.json", "r") as fp:
            self.result = json.load(fp)
            
    def plot_predictions(
            self, data_variable: str, x_dim: str, ax=None, subset={}, 
            upscale_x=True
        ):
        obs = self.simulation.observations.sel(subset)
        x_old = self.simulation.coordinates[x_dim].copy()
        
        nan_obs = obs[data_variable].isnull().all(dim=x_dim)

        if x_dim is not None:    
            if upscale_x:
                self.simulation.coordinates[x_dim] = np.linspace(
                    x_old.min(),
                    x_old.max(),
                    1000
                )
            else:
                raise NotImplementedError("x_new must be a 1D np.ndarray")
            
        Y = self.simulation.evaluate(self.result["X"])
        results = self.simulation.results_to_df(results=Y)

        if ax is None:
            ax = plt.subplot(111)
        
        ax.plot(
            results[x_dim].values, 
            results[data_variable].values.T[:, ~nan_obs], 
            color="black", lw=.8
        )

        ax.plot(
            obs[x_dim].values, 
            obs[data_variable].values[:, ~nan_obs], 
            marker="o", ls="", ms=3
        )
        
        ax.set_ylabel(data_variable)
        ax.set_xlabel(x_dim)

        # restore old coordinates
        self.simulation.coordinates[x_dim] = x_old

        return ax

if pymoo is not None:
    class OptimizationProblem(Problem):
        def __init__(self, backend: PymooBackend, **kwargs):
            self.backend = backend
            self.simulation = backend.simulation
            
            super().__init__(
                n_var=self.simulation.n_free_parameters, 
                n_obj=self.simulation.n_objectives, 
                n_ieq_constr=0, 
                xl=0.0, 
                xu=1.0,
                elementwise_evaluation=False,
                **kwargs
            )

        def _evaluate(self, X, out, *args, **kwargs):
            X_original_scale = self.backend.transform(X_scaled=X)       

            if self.backend.pool is None:
                F = list(map(self.backend.distance_function, X_original_scale))
            else:
                F = self.backend.pool.map(self.backend.distance_function, X_original_scale)

            out["F"] = np.array(F)


from . import parameters
from . import sections
from . import base
from . import casestudy_registry

from .sections import *

from .parameters import *

from .base import *

from .casestudy_registry import *


import os
import sys
import configparser
import warnings
import importlib
import importlib.util
from typing import List, Optional, Union, Dict, Literal, Tuple, Any
from types import ModuleType
import click
from pydantic import BaseModel, ConfigDict, Field
from pymob.sim.config.sections import (
    Casestudy,
    Simulation,
    Datastructure,
    Solverbase,
    Jaxsolver,
    Inference,
    Modelparameters,
    Errormodel,
    Multiprocessing,
    Pyabc,
    Pymoo,
    Numpyro,
    Redis,
    Report
)


from pymob.sim.config.casestudy_registry import get_case_study_model, _registry
from pymob.utils.store_file import scenario_file, converters
# this loads at the import of the module
default_path = sys.path.copy()

class Config(BaseModel):
    """
    Configuration manager for *pymob*.

    This class loads a ``settings.cfg`` file, validates it against the
    pydantic models defined above and provides convenient accessors.

    Parameters
    ----------
    config : str or configparser.ConfigParser, optional
        Path to a configuration file or an already parsed ``ConfigParser``.
    """
    model_config = ConfigDict(validate_assignment=True, extra="allow", protected_namespaces=())

    _config: configparser.ConfigParser = configparser.ConfigParser()
    _modules: Dict[str, ModuleType] = {}

    def __init__(
        self,
        config: Optional[Union[str, configparser.ConfigParser]] = None,
    ) -> None:

        _cfg_fp = None
        interp = configparser.ExtendedInterpolation()
        if isinstance(config, str):
            _config = configparser.ConfigParser(
                converters=converters,
                interpolation=interp
            )        
            _config.optionxform = str # type: ignore
            _cfg_file_paths = _config.read(config)
            try:
                _cfg_fp = _cfg_file_paths[0]
            except IndexError:
                raise FileNotFoundError(
                    f"Config file: {config} could not be found."
                ) 
        elif isinstance(config, configparser.ConfigParser):
            _config = config
        else:
            _config = configparser.ConfigParser(
                converters=converters,
                interpolation=interp
            )
            _config.optionxform = str # type: ignore

        # pass arguments to config

        if _cfg_fp is not None: 
            _config.set("case-study", "settings_path", _cfg_fp)
        cfg_dict = {k:dict(s) for k, s in dict(_config).items() if k != "DEFAULT"}
        
        # initalize case_study separately and import modules. This is done here,
        # so that the configuration sections from the registry are discoverable
        # this can only happen once the respective case studies are imported and with them
        # their settings models are defined registered (this of course must be done with the
        # register_case_study_config utility.

        # implementation sketch
        # TODO: Currently, everytime a Config is created (even with default configs)
        #       paths are added to sys path and this is printed. 
        #       by _import_casestudy_modules
        #       The easiest (hotfix) is to surpress these print statemetns when 
        #       the default settings are used.
        #       another (additional) opportunity would be to exploit this, if Config.__init__ is given 
        #       additional arguments (case-study name, scenario and package_location) 
        #       and a flag create_package.
        #       The arguments could alredy be used in the entry points simulate and infer
        #       the flat could be used in conjunction with _import_casestudy_modules 
        #       creating the package location, subdirectories and modules on the fly
        #       and importing them. This would be a nice jumpstarter for starting to work
        #       with case studies
        # if case_study_name is not None:
        #     cfg_dict["case_study"].update({"name": case_study_name})

        # if case_study_scenario is not None:
        #     cfg_dict["case_study"].update({"scenario": case_study_scenario})

        # if case_study_package is not None:
        #     cfg_dict["case_study"].update({"package": case_study_package})

        # if case_study_modules is not None:
        #     cfg_dict["case_study"].update({"modules": case_study_modules})

        _case_study_raw = cfg_dict.get("case-study", {})
        _case_study_instance = Casestudy.model_validate(_case_study_raw)
        _modules = self._import_casestudy_modules(case_study=_case_study_instance, reset_path=True)

        # initialize submodels
        super().__init__(**cfg_dict)

        self._config = _config
        self._modules = _modules

        # Load any case-study-specific configuration after generic sections are parsed
        self._load_case_study_section()

    case_study: Casestudy = Field(default=Casestudy(), alias="case-study")
    simulation: Simulation = Field(default=Simulation())
    data_structure: Datastructure = Field(default=Datastructure(), alias="data-structure") # type:ignore
    solverbase: Solverbase = Field(default=Solverbase())
    jaxsolver: Jaxsolver = Field(default=Jaxsolver(), alias="jax-solver")
    inference: Inference = Field(default=Inference())
    model_parameters: Modelparameters = Field(default=Modelparameters(), alias="model-parameters") #type: ignore
    error_model: Errormodel = Field(default=Errormodel(), alias="error-model") # type: ignore
    multiprocessing: Multiprocessing = Field(default=Multiprocessing())
    inference_pyabc: Pyabc = Field(default=Pyabc(), alias="inference.pyabc")
    inference_pyabc_redis: Redis = Field(default=Redis(), alias="inference.pyabc.redis")
    inference_pymoo: Pymoo = Field(default=Pymoo(), alias="inference.pymoo")
    inference_numpyro: Numpyro = Field(default=Numpyro(), alias="inference.numpyro")
    report: Report = Field(default=Report(), alias="report")
        
    @property
    def input_file_paths(self) -> List[str]:
        """
        List of all input files required for the simulation.

        Returns
        -------
        List[str]
        """
        paths_input_files = []
        for file in self.simulation.input_files:
            fp = scenario_file(file, self.case_study.name, self.case_study.scenario, pkg_dir=self.case_study.package)
            paths_input_files.append(fp)


        file = self.case_study.observations
        if file is None:
            return paths_input_files
        else:
            if not os.path.isabs(file):
                fp = os.path.join(self.case_study.data_path, file)
            else:
                fp = file
            paths_input_files.append(fp)

            return paths_input_files

    def print(self) -> None:
        """Print a summary of the configuration."""
        print("Simulation configuration", end="\n")
        print("========================")
        for section, field_info in self.model_fields.items():
            print(f"{section}({getattr(self, section)})", end="\n") # type: ignore

        print("========================", end="\n")

    def save(self, fp: Optional[str]=None, force: bool=False) -> None:
        """
        Save the configuration to a ``settings.cfg`` file.

        Uses serializers defined at the top, which parse the options to str
        so they can be processed by configfile. 

        In case the model configuration should be stored to a json file use
        something like `json.dumps(self.model_dump())`, because the build in
        function, is somewhat disabled by the listparsers which are needed for
        configfile lists.

        Parameters
        ----------
        fp : Optional[str]
            File path to write the settings file to. If ``None`` the default
            location derived from the case study is used.
        force : bool, optional
            Overwrite without prompting. Default is ``False``.
        """
        settings = self.model_dump(
            by_alias=True, 
            mode="json", 
            exclude_none=True,
            exclude={"case_study": {"output_path", "data_path", "root", "init_root", "default_settings_path"}}
        )
        self._config.update(**settings)

        if fp is None:
            file_path = self.case_study.default_settings_path
        else:
            file_path = os.path.abspath(fp)

        write = True
        if os.path.exists(file_path) and not force:
            ui_overwrite = input("Settings file already exists. Overwrite? [y/N]")
            write = True if ui_overwrite == "y" else False
        else:
            # create a directory for the new scenario file
            os.makedirs(os.path.dirname(file_path), exist_ok=True)

        if write:
            with open(file_path, "w") as f:
                self._config.write(f)

    def create_directory(self, directory: Literal["results", "scenario"], force: bool=False) -> None:
        """
        Create a results or scenario directory if it does not exist.

        Parameters
        ----------
        directory : Literal["results", "scenario"]
            Which directory to create.
        force : bool, optional
            If ``True`` create without prompting.
        """
        if directory == "results":
            p = os.path.abspath(self.case_study.output_path)
        elif directory == "scenario":
            p = os.path.abspath(self.case_study.scenario_path)
        else:
            raise NotImplementedError(
                f"{directory.capitalize()} is not an expected directory in the " +
                "case study logic. Use one of 'results' and 'scenario'."
            )

        if os.path.exists(p):
            print(f"{directory.capitalize()} directory exists at '{p}'.")
            return
        else:
            if not force:
                answer = input(f"Create {directory} directory at '{p}'? [Y/n]")
            else:
                answer = "y"

            if answer.lower() == "y" or answer.lower() == "":
                os.makedirs(p)
                print(f"{directory.capitalize()} directory created at '{p}'.")
            else:
                print(f"No {directory} directory created.")

    def import_casestudy_modules(self, reset_path: bool=False) -> None:
        """
        Import all modules of the current case study.

        this script handles the import of a case study without the typical 
        __init__.py file. It iterates over all .py files in the root directory
        of the case study (typically: sim, mod, stats, plot, data, prior)
        and imports them with import_module(...)

        Parameters
        ----------
        reset_path : bool, optional
            Reset ``sys.path`` before importing. Default is ``False``.

        """
        modules = self._import_casestudy_modules(
            case_study=self.case_study, 
            reset_path=reset_path
        )
        
        self._modules.update(modules)

    @staticmethod
    def _import_casestudy_modules(case_study: Casestudy, reset_path: bool=False) -> Dict[str, ModuleType]:
        """
        Import modules for a given case study.

        Parameters
        ----------
        case_study : Casestudy
            The case-study configuration.
        reset_path : bool, optional
            Whether to reset ``sys.path`` before importing.

        Returns
        -------
        Dict[str, ModuleType]
            Mapping of module name to imported module.
        """
        _modules = {}


        # reset the path to avoid importing modules form case-studies used
        # before in the same session
        if reset_path:
            # default path needs be copied, otherwise it will be updated
            # when setting sys.path
            sys.path = default_path.copy()


        # potential BUG: This is not safe. It is not guaranteed that the 
        # case study has the same name as the package. But it might be in the future
        package = case_study.name

        if "-" in package or " " in package:
            warnings.warn(
                f"Case-study contained {package} contained unallowed "+
                "characters: ['-', ' ']. "+
                "The characters will be replaced with underscores ('_') for "+
                "importing the package modules. " +
                "In the future, the name of the case study should be the same as " +
                "as the package where the modules are located. This name must not " +
                "contain hyphens ('-') or whitespace characters (' '),",
                category=UserWarning
            )
            _package = package.replace("-", "_").replace(" ", "_")
        else:
            _package = package

        spec = importlib.util.find_spec(_package)
        if spec is not None:
            for module in case_study.modules:
                try:
                    # TODO: Consider importing modules as a nested dictionary 
                    # with the indexing key being the package. The package
                    # cannot be derived from the class, if a method, that is 
                    # executed on a lower level case-study, should target that 
                    # a module belonging to the same package, because if the
                    # object is used, it would resolve to the package of the
                    # higher level case-study
                    m = importlib.import_module(f"{_package}.{module}")
                    _modules.update({module: m})
                except ModuleNotFoundError:
                    warnings.warn(
                        f"Module {module}.py not found in {_package}." +
                        "Missing modules can lead to unexpected behavior. " +
                        f"Does your case study have a {module}.py file? " +
                        "It should have the line `from PARENT_CASE_STUDY." +
                        f"{module} import *` to import all objects from " +
                        "the parent case study."
                    )
            return _modules

        # append relevant paths to sys
        package = os.path.join(
            case_study.root, 
            case_study.package
        )
        if package not in sys.path:
            sys.path.insert(0, package)
            print(f"Inserted '{package}' in PATH at index=0")
    
        case_study_path = os.path.join(
            case_study.root, 
            case_study.package,
            case_study.name,
            # Account for package architecture 
            case_study.name
        )
        if case_study_path not in sys.path:
            sys.path.insert(0, case_study_path)
            print(f"Inserted '{case_study_path}' in PATH at index=0")

        for module in case_study.modules:
            # remove modules of a different case study that might have been
            # loaded in the same session.
            if module in sys.modules:
                _ = sys.modules.pop(module)

        for module in case_study.modules:
            try:
                m = importlib.import_module(module, package=case_study_path)
                _modules.update({module: m})
            except ModuleNotFoundError:
                warnings.warn(
                    f"Module {module}.py not found in {case_study_path}." +
                    "Missing modules can lead to unexpected behavior." +
                    "If a module is not imported, you can specify it in the " +
                    "Config 'config.case_study.modules = [...]'"
                )

        return _modules

    def import_simulation_from_case_study(self) -> Any:
        """
        Retrieve the ``Simulation`` class defined in the case-study.

        Returns
        -------
        Any
            The ``Simulation`` class object.

        Raises
        ------
        ImportError
            If the class cannot be found.
        """
        try:
            Simulation = getattr(self._modules["sim"], self.case_study.simulation)
        except Exception as e:
            raise ImportError(
                f"Simulation class '{self.case_study.simulation}' " +
                "could not be found. Make sure the simulaton option is spelled " +
                "correctly or specify a class that exists in sim.py" +
                "If you are using pymob to work on different case-studies in " +
                "the same session, make sure to reset the path by " +
                "using `import_casestudy_modules(reset_path=True)`" +
                f"\nOriginal exception: {e}"
            )
        
        return Simulation

    def set_option(self, section: str, option: str, value: str) -> None:
        """
        Set a configuration option.

        Parameters
        ----------
        section : str
            Name of the configuration section (e.g. ``simulation``).
        option : str
            Option name within the section.
        value : str
            New value as a string; will be parsed according to the section's
            type definitions.
        """
        sect = getattr(self, section)
        if isinstance(sect, Modelparameters):
            if (value == "" or value == "None") and option in sect.all:
                sect.remove(option)
            elif (value == "" or value == "None") and option not in sect.all:
                pass
            else:
                sect[option] = value
        else:
            sect[option] = value

    def _load_case_study_section(self) -> None:
        """
        Parse a registered case-study-specific INI section and expose it as an attribute.
        """
        # Ensure the model has extra sections
        if self.model_extra is None:
            return

        # parse the config sections if they are defined in the settings.cfg
        # this overwrites the default definitions
        for section, options in self.model_extra.items():
            # get the model from the registry
            model_cls = get_case_study_model(section)

            if model_cls is None:
                warnings.warn(
                    f"Config section {section} could not be parsed. " +
                    "You need to register a the Config Model like that:\n\n" +
                    ">>> from pymob.sim.casestudy_registry import register_case_study_config\n" +
                    ">>> from pymob.sim.config import PymobModel\n" +
                    f">>> class {section.capitalize()}Config(PymobModel):\n" +
                    ">>>     option_a: float = 1.0\n" +
                    f">>> register_case_study_config({section.capitalize()}Config)\n"
                )

            else:
                # Pydantic parses/validates the raw strings
                instance = model_cls.model_validate(options, strict=False)
                # Store on the Config object – attribute name equals the case-study name
                self.model_extra.update({section: instance})

        # add default sections in case the sections are not defined in the settings.cfg
        for section, model_cls in _registry.items():
            if section not in self.model_extra:
                default_instance = model_cls.model_validate({}, strict=False)
                self.model_extra.update({section: default_instance})
            

@click.command
@click.option("--file", "-f", type=str, nargs=1, help="Path to the config file (usually in scenario/.../settings.cfg)")
@click.option("--options", "-o", type=str, multiple=True, help="The option or options to configure. Combine sections and option like this 'simulation.seed=1' options that have spaces need to be wraped in quotes")
def configure(file: str, options: Tuple[str, ...]) -> None:
    """
    Command-line entry point to modify a configuration file.

    Parameters
    ----------
    file : str
        Path to the configuration file.
    options : Tuple[str, ...]
        List of ``section.option=value`` strings.

    Examples
    --------
    >>> # Change the seed of the simulation
    >>> configure("-f", "scenario/settings.cfg", "-o", "simulation.seed=42")
    """
    config = Config(file)
    for opt in options:
        key, val = opt.split("=", 1)

        section, option = key.strip(" ").rsplit(".", 1)

        section = section.replace("-","_").replace(".","_")

        if section == "jax_solver":
            section = "jaxsolver"

        value = val.strip(" ")
        config.set_option(section, option, value)

    config.save(file, force=True)
